{
 "cells": [
  {
   "cell_type": "markdown",
   "id": "429e9a06",
   "metadata": {},
   "source": [
    "# Outlook\n",
    "\n",
    "In this notebook, using BBRL, we code the DDPG algorithm.\n",
    "\n",
    "To understand this code, you need to know more about \n",
    "[the BBRL interaction model](https://github.com/osigaud/bbrl/blob/master/docs/overview.md)\n",
    "Then you should run [a didactical example](https://github.com/osigaud/bbrl/blob/master/docs/notebooks/03-multi_env_autoreset.student.ipynb)\n",
    "to see how agents interact in BBRL when autoreset=True.\n",
    "\n",
    "The DDPG algorithm is explained in [this video](https://www.youtube.com/watch?v=0D6a0a1HTtc)\n",
    "and you can also read [the # corresponding slides](http://pages.isir.upmc.fr/~sigaud/teach/ddpg.pdf)."
   ]
  },
  {
   "cell_type": "markdown",
   "id": "53150fc4",
   "metadata": {},
   "source": [
    "## Installation and Imports\n",
    "\n",
    "### Installation\n",
    "\n",
    "The BBRL library is [here](https://github.com/osigaud/bbrl).\n",
    "\n",
    "We use OmegaConf to that makes it possible that by just defining the `def\n",
    "run_dqn(cfg):` function and then executing a long `params = {...}` variable at\n",
    "the bottom of this colab, the code is run with the parameters without calling\n",
    "an explicit main.\n",
    "\n",
    "More precisely, the code is run by calling\n",
    "\n",
    "`config=OmegaConf.create(params)`\n",
    "\n",
    "`run_dqn(config)`\n",
    "\n",
    "at the very bottom of the colab, after starting tensorboard.\n",
    "\n",
    "Below, we import standard python packages, pytorch packages and gymnasium\n",
    "environments."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "fce89d23",
   "metadata": {
    "jupyter": {
     "source_hidden": true
    },
    "tags": [
     "hide-input"
    ]
   },
   "outputs": [],
   "source": [
    "# Installs the necessary Python and system libraries\n",
    "try:\n",
    "    from easypip import easyimport, easyinstall, is_notebook\n",
    "except ModuleNotFoundError as e:\n",
    "    get_ipython().run_line_magic(\"pip\", \"install easypip\")\n",
    "    from easypip import easyimport, easyinstall, is_notebook\n",
    "\n",
    "easyinstall(\"bbrl>=0.2.2\")\n",
    "easyinstall(\"swig\")\n",
    "easyinstall(\"bbrl_gymnasium>=0.2.0\")\n",
    "easyinstall(\"bbrl_gymnasium[box2d]\")\n",
    "easyinstall(\"bbrl_gymnasium[classic_control]\")\n",
    "easyinstall(\"tensorboard\")\n",
    "easyinstall(\"moviepy\")\n",
    "easyinstall(\"box2d-kengz\")"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "62ab7772",
   "metadata": {
    "jupyter": {
     "source_hidden": true
    },
    "tags": [
     "hide-input"
    ]
   },
   "outputs": [],
   "source": [
    "import os\n",
    "import sys\n",
    "from pathlib import Path\n",
    "import math\n",
    "\n",
    "from moviepy.editor import ipython_display as video_display\n",
    "import time\n",
    "from tqdm.auto import tqdm\n",
    "from typing import Tuple, Optional\n",
    "from functools import partial\n",
    "\n",
    "from omegaconf import OmegaConf\n",
    "import torch\n",
    "import bbrl_gymnasium\n",
    "\n",
    "import copy\n",
    "from abc import abstractmethod, ABC\n",
    "import torch.nn as nn\n",
    "import torch.nn.functional as F\n",
    "from time import strftime\n",
    "\n",
    "OmegaConf.register_new_resolver(\n",
    "    \"current_time\", lambda: strftime(\"%Y%m%d-%H%M%S\"), replace=True\n",
    ")"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "e289997c",
   "metadata": {
    "jupyter": {
     "source_hidden": true
    },
    "lines_to_next_cell": 2,
    "tags": [
     "hide-input"
    ]
   },
   "outputs": [],
   "source": [
    "# Imports all the necessary classes and functions from BBRL\n",
    "from bbrl.agents.agent import Agent\n",
    "from bbrl import get_arguments, get_class, instantiate_class\n",
    "\n",
    "# The workspace is the main class in BBRL, this is where all data is collected and stored\n",
    "from bbrl.workspace import Workspace\n",
    "\n",
    "# Agents(agent1,agent2,agent3,...) executes the different agents the one after the other\n",
    "# TemporalAgent(agent) executes an agent over multiple timesteps in the workspace,\n",
    "# or until a given condition is reached\n",
    "from bbrl.agents import Agents, TemporalAgent\n",
    "\n",
    "# ParallelGymAgent is an agent able to execute a batch of gymnasium environments\n",
    "# with auto-resetting. These agents produce multiple variables in the workspace:\n",
    "# ’env/env_obs’, ’env/reward’, ’env/timestep’, ’env/terminated’,\n",
    "# 'env/truncated', 'env/done', ’env/cumulated_reward’, ...\n",
    "#\n",
    "# When called at timestep t=0, the environments are automatically reset. At\n",
    "# timestep t>0, these agents will read the ’action’ variable in the workspace at\n",
    "# time t − 1\n",
    "from bbrl.agents.gymnasium import GymAgent, ParallelGymAgent, make_env, record_video\n",
    "\n",
    "# Replay buffers are useful to store past transitions when training\n",
    "from bbrl.utils.replay_buffer import ReplayBuffer"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "3b30a039",
   "metadata": {
    "jupyter": {
     "source_hidden": true
    },
    "tags": [
     "hide-input"
    ]
   },
   "outputs": [],
   "source": [
    "# Utility function for launching tensorboard\n",
    "# For Colab - otherwise, it is easier and better to launch tensorboard from\n",
    "# the terminal\n",
    "def setup_tensorboard(path):\n",
    "    path = Path(path)\n",
    "    answer = \"\"\n",
    "    if is_notebook():\n",
    "        if get_ipython().__class__.__module__ == \"google.colab._shell\":\n",
    "            answer = \"y\"\n",
    "        while answer not in [\"y\", \"n\"]:\n",
    "            answer = input(\n",
    "                f\"Do you want to launch tensorboard in this notebook [y/n] \"\n",
    "            ).lower()\n",
    "\n",
    "    if answer == \"y\":\n",
    "        get_ipython().run_line_magic(\"load_ext\", \"tensorboard\")\n",
    "        get_ipython().run_line_magic(\"tensorboard\", f\"--logdir {path.absolute()}\")\n",
    "    else:\n",
    "        import sys\n",
    "        import os\n",
    "        import os.path as osp\n",
    "\n",
    "        print(\n",
    "            f\"Launch tensorboard from the shell:\\n{osp.dirname(sys.executable)}/tensorboard --logdir={path.absolute()}\"\n",
    "        )"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "06d7c0df",
   "metadata": {},
   "source": [
    "The [DDPG](https://arxiv.org/pdf/1509.02971.pdf) algorithm is an actor critic\n",
    "algorithm. We use a simple neural network builder function. This neural\n",
    "networks plays the role of the actor and the critic."
   ]
  },
  {
   "cell_type": "markdown",
   "id": "ef3ae162",
   "metadata": {},
   "source": [
    "## Definition of agents"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "7336c4aa",
   "metadata": {},
   "source": [
    "The function below builds a multi-layer perceptron where the size of each layer is given in the `size` list.\n",
    "We also specify the activation function of neurons at each layer and optionally a different activation function for the final layer."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "174f588c",
   "metadata": {},
   "outputs": [],
   "source": [
    "import torch.nn as nn\n",
    "def build_mlp(sizes, activation, output_activation=nn.Identity()):\n",
    "    \"\"\"Helper function to build a multi-layer perceptron (function from $\\mathbb R^n$ to $\\mathbb R^p$)\n",
    "    \n",
    "    Args:\n",
    "        sizes (List[int]): the number of neurons at each layer\n",
    "        activation (nn.Module): a PyTorch activation function (after each layer but the last)\n",
    "        output_activation (nn.Module): a PyTorch activation function (last layer)\n",
    "    \"\"\"\n",
    "    layers = []\n",
    "    for j in range(len(sizes) - 1):\n",
    "        act = activation if j < len(sizes) - 2 else output_activation\n",
    "        layers += [nn.Linear(sizes[j], sizes[j + 1]), act]\n",
    "    return nn.Sequential(*layers)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "c148c5a8",
   "metadata": {},
   "source": [
    "The critic is a neural network taking the state $s$ and action $a$ as input,\n",
    "and its output layer has a unique neuron whose value is the value of being in\n",
    "that state and performing that action $Q(s,a)$.\n",
    "\n",
    "As usual, the ```forward(...)``` function is used to write Q-values in the\n",
    "workspace from time indexes, whereas the ```predict_value(...)``` function` is\n",
    "used in other contexts, such as plotting a view of the Q function."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "c1b4b93f",
   "metadata": {
    "lines_to_next_cell": 2
   },
   "outputs": [],
   "source": [
    "class ContinuousQAgent(Agent):\n",
    "    def __init__(self, state_dim, hidden_layers, action_dim):\n",
    "        super().__init__()\n",
    "        self.is_q_function = True\n",
    "        self.model = build_mlp(\n",
    "            [state_dim + action_dim] + list(hidden_layers) + [1], activation=nn.ReLU()\n",
    "        )\n",
    "\n",
    "\n",
    "    def forward(self, t):\n",
    "        # Get the current state $s_t$ and the chosen action $a_t$\n",
    "        obs = self.get((\"env/env_obs\", t)) # shape B x D_{obs}\n",
    "        action = self.get((\"action\", t))   # shape B x D_{action}\n",
    "\n",
    "        # Compute the Q-value(s_t, a_t)\n",
    "        obs_act = torch.cat((obs, action), dim=1)  # shape B x (D_{obs} + D_{action})\n",
    "        # Get the q-value (and remove the last dimension since it is a scalar)\n",
    "        q_value = self.model(obs_act).squeeze(-1)\n",
    "        self.set((f\"{self.prefix}q_value\", t), q_value)\n",
    "\n",
    "    def predict_value(self, obs, action):\n",
    "        obs_act = torch.cat((obs, action), dim=0)\n",
    "        q_value = self.model(obs_act)\n",
    "        return q_value"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "20b38cb5",
   "metadata": {
    "lines_to_next_cell": 2
   },
   "source": [
    "The actor is also a neural network, it takes a state $s$ as input and outputs\n",
    "an action $a$."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "d8558a10",
   "metadata": {
    "lines_to_next_cell": 2
   },
   "outputs": [],
   "source": [
    "class ContinuousDeterministicActor(Agent):\n",
    "    def __init__(self, state_dim, hidden_layers, action_dim):\n",
    "        super().__init__()\n",
    "        layers = [state_dim] + list(hidden_layers) + [action_dim]\n",
    "        self.model = build_mlp(\n",
    "            layers, activation=nn.ReLU(), output_activation=nn.Tanh()\n",
    "        )\n",
    "\n",
    "    def forward(self, t, **kwargs):\n",
    "        obs = self.get((\"env/env_obs\", t))\n",
    "        action = self.model(obs)\n",
    "        self.set((\"action\", t), action)\n",
    "\n",
    "    def predict_action(self, obs, stochastic):\n",
    "        assert (\n",
    "            not stochastic\n",
    "        ), \"ContinuousDeterministicActor cannot provide stochastic predictions\"\n",
    "        return self.model(obs)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "8fc691fc",
   "metadata": {
    "lines_to_next_cell": 2
   },
   "source": [
    "### Creating an Exploration method\n",
    "\n",
    "In the continuous action domain, basic exploration differs from the methods\n",
    "used in the discrete action domain. Here we generally add some Gaussian noise\n",
    "to the output of the actor."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "b9c911ec",
   "metadata": {},
   "outputs": [],
   "source": [
    "from torch.distributions import Normal"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "f65ec96f",
   "metadata": {},
   "outputs": [],
   "source": [
    "class AddGaussianNoise(Agent):\n",
    "    def __init__(self, sigma):\n",
    "        super().__init__()\n",
    "        self.sigma = sigma\n",
    "\n",
    "    def forward(self, t, **kwargs):\n",
    "        act = self.get((\"action\", t))\n",
    "        dist = Normal(act, self.sigma)\n",
    "        action = dist.sample()\n",
    "        self.set((\"action\", t), action)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "ba822020",
   "metadata": {},
   "source": [
    "In [the original DDPG paper](https://arxiv.org/pdf/1509.02971.pdf), the\n",
    "authors rather used the more sophisticated Ornstein-Uhlenbeck noise where\n",
    "noise is correlated between one step and the next."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "7f2ff13c",
   "metadata": {},
   "outputs": [],
   "source": [
    "class AddOUNoise(Agent):\n",
    "    \"\"\"\n",
    "    Ornstein Uhlenbeck process noise for actions as suggested by DDPG paper\n",
    "    \"\"\"\n",
    "\n",
    "    def __init__(self, std_dev, theta=0.15, dt=1e-2):\n",
    "        self.theta = theta\n",
    "        self.std_dev = std_dev\n",
    "        self.dt = dt\n",
    "        self.x_prev = 0\n",
    "\n",
    "    def forward(self, t, **kwargs):\n",
    "        act = self.get((\"action\", t))\n",
    "        x = (\n",
    "            self.x_prev\n",
    "            + self.theta * (act - self.x_prev) * self.dt\n",
    "            + self.std_dev * math.sqrt(self.dt) * torch.randn(act.shape)\n",
    "        )\n",
    "        self.x_prev = x\n",
    "        self.set((\"action\", t), x)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "b3041938",
   "metadata": {},
   "source": [
    "### The Logger class\n",
    "\n",
    "The logger is in charge of collecting statistics during the training\n",
    "process.\n",
    "\n",
    "Having logging provided under the hood is one of the features allowing you\n",
    "to save time when using RL libraries like BBRL.\n",
    "\n",
    "In these notebooks, the logger is defined as `bbrl.utils.logger.TFLogger` so as\n",
    "to use a tensorboard visualisation (see the parameters part `params = { \"logger\":{ ...` below).\n",
    "\n",
    "Note that the BBRL Logger is also saving the log in a readable format such\n",
    "that you can use `Logger.read_directories(...)` to read multiple logs, create\n",
    "a dataframe, and analyze many experiments afterward in a notebook for\n",
    "instance. The code for the different kinds of loggers is available in the\n",
    "[bbrl/utils/logger.py](https://github.com/osigaud/bbrl/blob/master/src/bbrl/utils/logger.py)\n",
    "file.\n",
    "\n",
    "`instantiate_class` is an inner BBRL mechanism. The\n",
    "`instantiate_class`function is available in the\n",
    "[`bbrl/__init__.py`](https://github.com/osigaud/bbrl/blob/master/src/bbrl/__init__.py)\n",
    "file."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "0ae7138d",
   "metadata": {
    "jupyter": {
     "source_hidden": true
    },
    "lines_to_next_cell": 2,
    "tags": [
     "hide-input"
    ]
   },
   "outputs": [],
   "source": [
    "from bbrl import instantiate_class\n",
    "\n",
    "class Logger():\n",
    "\n",
    "    def __init__(self, cfg):\n",
    "        self.logger = instantiate_class(cfg.logger)\n",
    "\n",
    "    def add_log(self, log_string, loss, steps):\n",
    "        self.logger.add_scalar(log_string, loss.item(), steps)\n",
    "\n",
    "    # A specific function for RL algorithms having a critic, an actor and an entropy losses\n",
    "    def log_losses(self, critic_loss, entropy_loss, actor_loss, steps):\n",
    "        self.add_log(\"critic_loss\", critic_loss, steps)\n",
    "        self.add_log(\"entropy_loss\", entropy_loss, steps)\n",
    "        self.add_log(\"actor_loss\", actor_loss, steps)\n",
    "\n",
    "    def log_reward_losses(self, rewards, nb_steps):\n",
    "        self.add_log(\"reward/mean\", rewards.mean(), nb_steps)\n",
    "        self.add_log(\"reward/max\", rewards.max(), nb_steps)\n",
    "        self.add_log(\"reward/min\", rewards.min(), nb_steps)\n",
    "        self.add_log(\"reward/median\", rewards.median(), nb_steps)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "b9aff055",
   "metadata": {},
   "source": [
    "### Compute critic loss\n",
    "\n",
    "Detailed explanations of the function to compute the critic loss\n",
    "are given in [the DQN notebook](http://master-dac.isir.upmc.fr/rld/rl/03-1-dqn-introduction.student.ipynb)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "5b9ab4a4",
   "metadata": {
    "lines_to_next_cell": 1
   },
   "outputs": [],
   "source": [
    "mse = nn.MSELoss()\n",
    "\n",
    "def compute_critic_loss(cfg, reward, must_bootstrap, q_values, target_q_values):\n",
    "    \"\"\"Compute the DDPG critic loss from a sample of transitions\n",
    "\n",
    "    :param cfg: The configuration\n",
    "    :param reward: The reward (shape 2xB)\n",
    "    :param must_bootstrap: Must bootstrap flag (shape 2xB)\n",
    "    :param q_values: The computed Q-values (shape 2xB)\n",
    "    :param target_q_values: The Q-values computed by the target critic (shape 2xB)\n",
    "    :return: the loss (a scalar)\n",
    "    \"\"\"    \n",
    "    # Compute temporal difference\n",
    "    target = (\n",
    "        reward[1]\n",
    "        + cfg.algorithm.discount_factor * target_q_values[1] * must_bootstrap[1].int()\n",
    "    )\n",
    "    # Compute critic loss\n",
    "    critic_loss = mse(q_values[0], target)\n",
    "    return critic_loss"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "fb386df9",
   "metadata": {
    "lines_to_next_cell": 2
   },
   "source": [
    "### Soft parameter updates\n",
    "\n",
    "To update the target critic, one uses the following equation:\n",
    "$\\theta' \\leftarrow \\tau \\theta + (1- \\tau) \\theta'$\n",
    "where $\\theta$ is the vector of parameters of the critic, and $\\theta'$ is the vector of parameters of the target critic.\n",
    "The `soft_update_params(...)` function is in charge of performing this soft update."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "23080fea",
   "metadata": {
    "jupyter": {
     "source_hidden": true
    },
    "tags": [
     "hide-input"
    ]
   },
   "outputs": [],
   "source": [
    "def soft_update_params(net, target_net, tau):\n",
    "    for param, target_param in zip(net.parameters(), target_net.parameters()):\n",
    "        target_param.data.copy_(tau * param.data + (1 - tau) * target_param.data)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "a4e3fe42",
   "metadata": {},
   "source": [
    "### Compute actor loss\n",
    "The actor loss is straightforward. We want the actor to maximize Q-values, thus we minimize the mean of negated Q-values."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "47715f73",
   "metadata": {},
   "outputs": [],
   "source": [
    "def compute_actor_loss(q_values):\n",
    "    \"\"\"Returns the actor loss\n",
    "\n",
    "    :param q_values: The q-values (shape 2xB)\n",
    "    :return: A scalar (the loss)\n",
    "    \"\"\"\n",
    "    return -q_values[0].mean()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "ef1e92f3",
   "metadata": {},
   "outputs": [],
   "source": [
    "from bbrl import get_arguments, get_class\n",
    "from itertools import chain\n",
    "\n",
    "def setup_optimizer(cfg_optimizer, *agents):\n",
    "    \"\"\"Setup an optimizer for a list of agents\"\"\"\n",
    "    optimizer_args = get_arguments(cfg_optimizer)\n",
    "    parameters = [agent.parameters() for agent in agents]\n",
    "    optimizer = get_class(cfg_optimizer)(chain(*parameters), **optimizer_args)\n",
    "    return optimizer\n",
    "\n",
    "def copy_parameters(model_a, model_b):\n",
    "    \"\"\"Copy parameters from a model a to model_b\"\"\"\n",
    "    for model_a_p, model_b_p in zip(model_a.parameters(), model_b.parameters()):\n",
    "        model_b_p.data.copy_(model_a_p)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "718fd87d",
   "metadata": {
    "lines_to_next_cell": 2
   },
   "source": [
    "# Learning environment\n",
    "\n",
    "To setup a common learning environment for RL algorithms, we use the `RLBase`\n",
    "class. This class:\n",
    "\n",
    "1. Initializes the environment (random seed, logger, evaluation environment)\n",
    "2. Defines a `evaluate` method that keeps the best agent so far\n",
    "3. Defines a `visualize_best` method that displays the behavior of the best agent\n",
    "\n",
    "Subclasses need to define `self.train_policy` and `self.eval_policy`, two\n",
    "BBRL agents that respectively choose actions when training and evaluating.\n",
    "\n",
    "The behavior of `RLBase` is controlled by the following configuration\n",
    "variables:\n",
    "\n",
    "- `base_dir` defines the directory subpath used when outputing losses during\n",
    "training as well as other outputs (serialized agent, global statistics, etc.)\n",
    "- `algorithm.seed` defines the random seed used (to initialize the agent and\n",
    "  the environment)\n",
    "- `gym_env` defines the gymnasium environment, and in particular\n",
    "`gym_env.env_name` the name of the gymansium environment\n",
    "- `logger` defines what type of logger is used to log the different values\n",
    "associated with learning\n",
    "- `algorithm.eval_interval` defines the number of observed transitions between\n",
    "each evaluation of the agent"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "2169b56b",
   "metadata": {
    "jupyter": {
     "source_hidden": true
    },
    "tags": [
     "hide-input"
    ]
   },
   "outputs": [],
   "source": [
    "import numpy as np\n",
    "from typing import Any\n",
    "import logging\n",
    "from abc import ABC\n",
    "from functools import cached_property\n",
    "\n",
    "class RLBase(ABC):\n",
    "    \"\"\"Base class for Reinforcement learning algorithms\n",
    "    \n",
    "    This class deals with common processing:\n",
    "\n",
    "    - defines the logger, the train and evaluation agents\n",
    "    - defines how to evaluate a policy\n",
    "    \"\"\"\n",
    "\n",
    "    #: The configuration\n",
    "    cfg: Any\n",
    "\n",
    "    #: The evaluation environment deals with the last action, and produces a new\n",
    "    # state of the environment\n",
    "    eval_env: Agent\n",
    "\n",
    "    #: The training policy\n",
    "    train_policy: Agent\n",
    "\n",
    "    #: The evaluation policy (if not defined, uses the training policy)\n",
    "    eval_policy: Agent\n",
    "\n",
    "    def __init__(self, cfg):\n",
    "        # Basic initialization\n",
    "        self.cfg = cfg\n",
    "        torch.manual_seed(cfg.algorithm.seed)\n",
    "\n",
    "        # Sets the base directory and logger directory\n",
    "        base_dir = Path(self.cfg.base_dir)\n",
    "        self.base_dir = Path(\"outputs\") / Path(self.cfg.base_dir)\n",
    "        self.base_dir.mkdir(parents=True, exist_ok=True)\n",
    "\n",
    "        # Initialize the logger class\n",
    "        if not hasattr(cfg.logger, \"log_dir\"):\n",
    "            cfg.logger.log_dir = str(Path(\"outputs\") / \"tblogs\" / base_dir)\n",
    "        self.logger = Logger(cfg)\n",
    "\n",
    "        # Subclasses have to define the training and eval policies\n",
    "        self.train_policy = None\n",
    "        self.eval_policy = None\n",
    "\n",
    "        # Sets up the evaluation environment\n",
    "        self.eval_env = ParallelGymAgent(\n",
    "            partial(make_env, cfg.gym_env.env_name), \n",
    "            cfg.algorithm.nb_evals\n",
    "        ).seed(cfg.algorithm.seed)\n",
    "\n",
    "        # Initialize values\n",
    "        self.last_eval_step = 0\n",
    "        self.nb_steps = 0\n",
    "        self.best_policy = None\n",
    "        self.best_reward = -torch.inf\n",
    "\n",
    "        # Records the rewards\n",
    "        self.eval_rewards = []\n",
    "\n",
    "    @cached_property\n",
    "    def train_agent(self):\n",
    "        \"\"\"Returns the training agent\n",
    "\n",
    "        The agent is composed of a policy agent and the training environment.      \n",
    "        This method supposes that `self.train_policy` has been setup\n",
    "        \"\"\"\n",
    "        assert self.train_policy is not None, \"The train_policy property is not defined before the policy is set\"\n",
    "        return TemporalAgent(Agents(self.train_env, self.train_policy))\n",
    "\n",
    "    @cached_property\n",
    "    def eval_agent(self):\n",
    "        \"\"\"Returns the evaluation agent \n",
    "        \n",
    "        The agent is composed of a policy agent and the evaluation environment\n",
    "        \n",
    "        Uses `self.eval_policy` (or `self.train_policy` if not defined)\n",
    "\n",
    "        \"\"\"\n",
    "        assert self.eval_policy is not None or self.train_policy is not None, \"eval_agent property is not defined before the policy is set\"\n",
    "        return TemporalAgent(Agents(self.eval_env, self.eval_policy if self.eval_policy is not None else self.train_policy))\n",
    "\n",
    "    def evaluate(self):\n",
    "        \"\"\"Evaluate the current policy `self.eval_policy`\n",
    "        \n",
    "        Evaluation is conducted every `cfg.algorithm.eval_interval` steps, and\n",
    "        we keep a copy of the best agent so far in `self.best_policy`\n",
    "        \n",
    "        Returns True if the current policy is the best so far\n",
    "        \"\"\"\n",
    "        if (self.nb_steps - self.last_eval_step) > self.cfg.algorithm.eval_interval:\n",
    "            self.last_eval_step = self.nb_steps\n",
    "            eval_workspace = Workspace() \n",
    "            self.eval_agent(\n",
    "                eval_workspace,\n",
    "                t=0,\n",
    "                stop_variable=\"env/done\"\n",
    "            )\n",
    "            rewards = eval_workspace[\"env/cumulated_reward\"][-1]\n",
    "            self.logger.log_reward_losses(rewards, self.nb_steps)\n",
    "\n",
    "            if getattr(self.cfg, \"collect_stats\", False):\n",
    "                self.eval_rewards.append(rewards)\n",
    "\n",
    "            rewards_mean = rewards.mean()\n",
    "            if rewards_mean > self.best_reward:\n",
    "                self.best_policy = copy.deepcopy(self.eval_policy)\n",
    "                self.best_reward = rewards_mean\n",
    "                return True\n",
    "\n",
    "    def save_stats(self):\n",
    "        \"\"\"Save reward statistics into `stats.npy`\"\"\"\n",
    "        if getattr(self.cfg, \"collect_stats\", False) and self.eval_rewards:\n",
    "            data = torch.stack(self.eval_rewards, axis=-1) \n",
    "            with (self.base_dir / \"stats.npy\").open(\"wt\") as fp:\n",
    "                np.savetxt(fp, data.numpy())\n",
    "\n",
    "    def visualize_best(self):\n",
    "        \"\"\"Visualize the best agent\"\"\"\n",
    "        env = make_env(self.cfg.gym_env.env_name, render_mode=\"rgb_array\")\n",
    "        path = self.base_dir / \"best_agent.mp4\"\n",
    "        print(f\"Video of best agent recorded in {path}\")\n",
    "        record_video(env, self.best_policy, path)\n",
    "        return video_display(str(path.absolute()))"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "b778a4b9",
   "metadata": {
    "lines_to_next_cell": 2
   },
   "source": [
    "The `EpochBasedAlgo` defines the environment when using replay buffers. In\n",
    "particular, it defines `self.train_env` which is the environment used for\n",
    "training. This environment uses *autoreset*, i.e. when reaching a terminal\n",
    "state, a new environment is created.\n",
    "\n",
    "`EpochBasedAlgo` stores all the transitions $(s_t, a_t, r_t, s_{t+1}, ...)$\n",
    "into `self.replay_buffer`.\n",
    "\n",
    "The behavior of `EpochBasedAlgo` is controlled by the following configuration\n",
    "variables:\n",
    "\n",
    "- `gym_env.env_name` defines the gymnasium environment\n",
    "- `algorithm.n_envs` defines the number of parallel environments\n",
    "- `algorithm.seed` defines the random seed used (to initialize the agent and\n",
    "  the environment)\n",
    "- `algorithm.buffer_size` is the maximum number of transitions stored into the\n",
    "replay buffer"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "89f5a34a",
   "metadata": {
    "jupyter": {
     "source_hidden": true
    },
    "tags": [
     "hide-input"
    ]
   },
   "outputs": [],
   "source": [
    "class EpochBasedAlgo(RLBase):\n",
    "    \"\"\"RL environment when using transition buffers\"\"\"\n",
    "    \n",
    "    train_agent: TemporalAgent\n",
    "    \n",
    "    \"\"\"Base class for RL experiments with full episodes\"\"\"\n",
    "    def __init__(self, cfg):\n",
    "        super().__init__(cfg)\n",
    "\n",
    "        # We use a non-autoreset workspace\n",
    "        self.train_env = ParallelGymAgent(\n",
    "            partial(make_env, cfg.gym_env.env_name, autoreset=True), \n",
    "            cfg.algorithm.n_envs\n",
    "        ).seed(cfg.algorithm.seed)\n",
    "\n",
    "        # Configure the workspace to the right dimension\n",
    "        # Note that no parameter is needed to create the workspace.\n",
    "        self.replay_buffer = ReplayBuffer(max_size=cfg.algorithm.buffer_size)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "80188328",
   "metadata": {},
   "source": [
    "`iter_replay_buffers` provides an easy access to the replay buffer when\n",
    "learning. Its behavior depends on several configuration values:\n",
    "\n",
    "- `cfg.algorithm.max_epochs` defines the number of times the agent is used to\n",
    "collect transitions\n",
    "- `cfg.algorithm.learning_starts` defines the number of transitions before\n",
    "learning starts\n",
    " \n",
    "Using `iter_replay_buffers` is simple:\n",
    "\n",
    "```py\n",
    "  class MyAlgo(EpochBasedAlgo):\n",
    "      def __init__(self, cfg):\n",
    "          super().__init__(cfg)\n",
    "\n",
    "          # Define the train and evaluation policies\n",
    "          # (the agents compute the workspace `action` variable)\n",
    "          self.train_policy = ...\n",
    "          self.eval_policy = ...\n",
    "\n",
    "  rl_algo = MyAlgo(cfg)\n",
    "  for rb in iter_replay_buffers(rl_algo):\n",
    "      # rb is a workspace containing transitions\n",
    "      ...\n",
    "```"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "12205542",
   "metadata": {
    "jupyter": {
     "source_hidden": true
    },
    "tags": [
     "hide-input"
    ]
   },
   "outputs": [],
   "source": [
    "def iter_replay_buffers(algo: EpochBasedAlgo):\n",
    "    \"\"\"Loop over transition buffers\"\"\"\n",
    "    train_workspace = Workspace()\n",
    "\n",
    "    epochs_pb = tqdm(range(algo.cfg.algorithm.max_epochs))\n",
    "    for epoch in epochs_pb:\n",
    "        \n",
    "        # This is the tricky part with transition buffers. The difficulty lies in the\n",
    "        # copy of the last step and the way to deal with the n_steps return.\n",
    "        #\n",
    "        # The call to `train_agent(workspace, t=1, n_steps=cfg.algorithm.n_timesteps -\n",
    "        # 1, stochastic=True)` makes the agent run a number of steps in the workspace.\n",
    "        # In practice, it calls the\n",
    "        # [`__call__(...)`](https://github.com/osigaud/bbrl/blob/master/src/bbrl/agents/agent.py#L59)\n",
    "        # function which makes a forward pass of the agent network using the workspace\n",
    "        # data and updates the workspace accordingly.\n",
    "        #\n",
    "        # Now, if we start at the first epoch (`epoch=0`), we start from the first step\n",
    "        # (`t=0`). But when subsequently we perform the next epochs (`epoch>0`), we must\n",
    "        # not forget to cover the transition at the border between the previous epoch\n",
    "        # and the current epoch. To avoid this risk, we copy the information from the\n",
    "        # last time step of the previous epoch into the first time step of the next\n",
    "        # epoch. This is explained in more details in [a previous\n",
    "        # notebook](https://colab.research.google.com/drive/1W9Y-3fa6LsPeR6cBC1vgwBjKfgMwZvP5).\n",
    "        if epoch == 0:\n",
    "            # First run: we start from scratch\n",
    "            algo.train_agent(\n",
    "                train_workspace, t=0, n_steps=algo.cfg.algorithm.n_steps, stochastic=True\n",
    "            )\n",
    "        else:\n",
    "            # Other runs: we copy the last step and start from there\n",
    "            train_workspace.zero_grad()\n",
    "            train_workspace.copy_n_last_steps(1)\n",
    "            algo.train_agent(\n",
    "                train_workspace, t=1, n_steps=algo.cfg.algorithm.n_steps-1, stochastic=True\n",
    "            )\n",
    "        \n",
    "        algo.nb_steps += algo.cfg.algorithm.n_steps * algo.cfg.algorithm.n_envs\n",
    "\n",
    "        # Add transitions to buffer\n",
    "        transition_workspace = train_workspace.get_transitions()\n",
    "        algo.replay_buffer.put(transition_workspace)\n",
    "        if algo.replay_buffer.size() > algo.cfg.algorithm.learning_starts:\n",
    "            yield algo.replay_buffer\n",
    "\n",
    "        # Eval\n",
    "        epochs_pb.set_description(\n",
    "            f\"nb_steps: {algo.nb_steps}, \"\n",
    "            f\"best reward: {algo.best_reward:.2f}\"\n",
    "        )"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "99e1cc32",
   "metadata": {},
   "source": [
    "### Create the DDPG agent\n",
    "\n",
    "In the next cell, we create the critic and the actor, but also an exploration\n",
    "agent to add noise and a target critic. The version below does not use a\n",
    "target actor as it proved hard to tune, but such a target actor is used in the\n",
    "original paper."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "5442c6e0",
   "metadata": {
    "lines_to_next_cell": 1
   },
   "outputs": [],
   "source": [
    "class DDPG(EpochBasedAlgo):\n",
    "    def __init__(self, cfg):\n",
    "        super().__init__(cfg)\n",
    "        \n",
    "        # we create the critic and the actor, but also an exploration agent to\n",
    "        # add noise and a target critic. The version below does not use a target\n",
    "        # actor as it proved hard to tune, but such a target actor is used in\n",
    "        # the original paper.\n",
    "        \n",
    "        obs_size, act_size = self.train_env.get_obs_and_actions_sizes()\n",
    "        self.critic = ContinuousQAgent(\n",
    "            obs_size, cfg.algorithm.architecture.critic_hidden_size, act_size\n",
    "        ).with_prefix(\"critic/\")\n",
    "        self.target_critic = copy.deepcopy(self.critic).with_prefix(\"target-critic/\")\n",
    "\n",
    "        self.actor = ContinuousDeterministicActor(\n",
    "            obs_size, cfg.algorithm.architecture.actor_hidden_size, act_size\n",
    "        )\n",
    "        \n",
    "        noise_agent = AddGaussianNoise(cfg.algorithm.action_noise) # alternative : AddOUNoise\n",
    "        \n",
    "        self.train_policy = Agents(self.actor, noise_agent)  \n",
    "        self.eval_policy = self.actor\n",
    "\n",
    "        # Define agents over time\n",
    "        self.t_actor = TemporalAgent(self.actor)\n",
    "        self.t_critic = TemporalAgent(self.critic)\n",
    "        self.t_target_critic = TemporalAgent(self.target_critic)\n",
    "\n",
    "        # Configure the optimizer\n",
    "        self.actor_optimizer = setup_optimizer(cfg.actor_optimizer, self.actor)\n",
    "        self.critic_optimizer = setup_optimizer(cfg.critic_optimizer, self.critic)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "27b3f7f3",
   "metadata": {},
   "source": [
    "## Main training loop\n",
    "\n",
    "In the following, we code the main loop "
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "e67a484c",
   "metadata": {},
   "outputs": [],
   "source": [
    "from bbrl.visu.plot_policies import plot_policy\n",
    "from bbrl.visu.plot_critics import plot_critic\n",
    "\n",
    "def run_ddpg(ddpg: DDPG):\n",
    "    for rb in iter_replay_buffers(ddpg):\n",
    "        rb_workspace = rb.get_shuffled(ddpg.cfg.algorithm.batch_size)\n",
    "        terminated, reward = rb_workspace[\n",
    "            \"env/terminated\", \"env/reward\"\n",
    "        ]\n",
    "\n",
    "        # Determines whether values of the critic should be propagated\n",
    "        # True if the episode reached a time limit or if the task was not done\n",
    "        # See https://github.com/osigaud/bbrl/blob/master/docs/time_limits.md\n",
    "        must_bootstrap = ~terminated\n",
    "\n",
    "        # Critic update\n",
    "        # compute q_values: at t, we have Q(s,a) from the (s,a) in the RB\n",
    "        ddpg.t_critic(rb_workspace, t=0, n_steps=1)\n",
    "\n",
    "        with torch.no_grad():\n",
    "            # replace the action at t+1 in the RB with \\pi(s_{t+1}), to compute\n",
    "            # Q(s_{t+1}, \\pi(s_{t+1}) below\n",
    "            ddpg.t_actor(rb_workspace, t=1, n_steps=1)\n",
    "            # compute q_values: at t+1 we have Q(s_{t+1}, \\pi(s_{t+1})\n",
    "            ddpg.t_target_critic(rb_workspace, t=1, n_steps=1)\n",
    "\n",
    "        # finally q_values contains the above collection at t=0 and t=1\n",
    "        q_values, post_q_values = rb_workspace[\"critic/q_value\", \"target-critic/q_value\"]\n",
    "\n",
    "        # Compute critic loss\n",
    "        critic_loss = compute_critic_loss(\n",
    "            ddpg.cfg, reward, must_bootstrap, q_values, post_q_values\n",
    "        )\n",
    "        ddpg.logger.add_log(\"critic_loss\", critic_loss, ddpg.nb_steps)\n",
    "        ddpg.critic_optimizer.zero_grad()\n",
    "        critic_loss.backward()\n",
    "        torch.nn.utils.clip_grad_norm_(\n",
    "            ddpg.critic.parameters(), ddpg.cfg.algorithm.max_grad_norm\n",
    "        )\n",
    "        ddpg.critic_optimizer.step()\n",
    "\n",
    "        # Actor update\n",
    "\n",
    "        # Now we determine the actions the current policy would take in the states from the RB\n",
    "        ddpg.t_actor(rb_workspace, t=0, n_steps=1)\n",
    "        \n",
    "        # We determine the Q values resulting from actions of the current policy\n",
    "        ddpg.t_critic(rb_workspace, t=0, n_steps=1)\n",
    "        \n",
    "        # and we back-propagate the corresponding loss to maximize the Q values\n",
    "        q_values = rb_workspace[\"critic/q_value\"]\n",
    "        actor_loss = compute_actor_loss(q_values)\n",
    "        \n",
    "        ddpg.logger.add_log(\"actor_loss\", actor_loss, ddpg.nb_steps)\n",
    "        \n",
    "        # if -25 < actor_loss < 0 and nb_steps > 2e5:\n",
    "        ddpg.actor_optimizer.zero_grad()\n",
    "        actor_loss.backward()\n",
    "        torch.nn.utils.clip_grad_norm_(\n",
    "            ddpg.actor.parameters(), ddpg.cfg.algorithm.max_grad_norm\n",
    "        )\n",
    "        ddpg.actor_optimizer.step()\n",
    "        \n",
    "        # Soft update of target q function\n",
    "        soft_update_params(ddpg.critic, ddpg.target_critic, ddpg.cfg.algorithm.tau_target)\n",
    "\n",
    "\n",
    "        if ddpg.evaluate():\n",
    "            if ddpg.cfg.plot_agents:\n",
    "                plot_policy(\n",
    "                    ddpg.actor,\n",
    "                    ddpg.eval_env,\n",
    "                    ddpg.best_reward,\n",
    "                    str(ddpg.base_dir / \"plots\"),\n",
    "                    ddpg.cfg.gym_env.env_name,\n",
    "                    stochastic=False,\n",
    "                )"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "a0e39d7d",
   "metadata": {},
   "source": [
    "# Definition of the parameters\n",
    "\n",
    "The logger is defined as `bbrl.utils.logger.TFLogger` so as to use a\n",
    "tensorboard visualisation."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "e224db2e",
   "metadata": {},
   "outputs": [],
   "source": [
    "params={\n",
    "  \"save_best\": False,\n",
    "  \"base_dir\": \"${gym_env.env_name}/ddpg-S${algorithm.seed}_${current_time:}\",\n",
    "  \"collect_stats\": True,\n",
    "  # Set to true to have an insight on the learned policy\n",
    "  # (but slows down the evaluation a lot!)\n",
    "  \"plot_agents\": True,\n",
    "  \"logger\":{\n",
    "    \"classname\": \"bbrl.utils.logger.TFLogger\",\n",
    "    \"cache_size\": 10000,\n",
    "    \"every_n_seconds\": 10,\n",
    "    \"verbose\": False,    \n",
    "    },\n",
    "\n",
    "  \"algorithm\":{\n",
    "    \"seed\": 1,\n",
    "    \"max_grad_norm\": 0.5,\n",
    "    \"epsilon\": 0.02,\n",
    "    \"n_envs\": 1,\n",
    "    \"n_steps\": 100,\n",
    "    \"nb_evals\": 10,\n",
    "    \"discount_factor\": 0.98,\n",
    "    \"buffer_size\": 2e5,\n",
    "    \"batch_size\": 64,\n",
    "    \"tau_target\": 0.05,\n",
    "    \"eval_interval\": 2_000,\n",
    "    \"max_epochs\": 11_000,\n",
    "    # Minimum number of transitions before learning starts\n",
    "    \"learning_starts\": 10000,\n",
    "    \"action_noise\": 0.1,\n",
    "    \"architecture\":{\n",
    "        \"actor_hidden_size\": [400, 300],\n",
    "        \"critic_hidden_size\": [400, 300],\n",
    "        },\n",
    "  },\n",
    "  \"gym_env\":{\n",
    "    \"env_name\": \"Pendulum-v1\",\n",
    "  },\n",
    "  \"actor_optimizer\":{\n",
    "    \"classname\": \"torch.optim.Adam\",\n",
    "    \"lr\": 1e-3,\n",
    "  },\n",
    "  \"critic_optimizer\":{\n",
    "    \"classname\": \"torch.optim.Adam\",\n",
    "    \"lr\": 1e-3,\n",
    "  }\n",
    "}"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "979b5d47",
   "metadata": {},
   "source": [
    "## Launching tensorboard to visualize the results"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "5139e03e",
   "metadata": {},
   "outputs": [],
   "source": [
    "setup_tensorboard(\"./outputs/tblogs\")"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "63dae8fb",
   "metadata": {
    "lines_to_next_cell": 1
   },
   "outputs": [],
   "source": [
    "ddpg = DDPG(OmegaConf.create(params))\n",
    "run_ddpg(ddpg)\n",
    "ddpg.visualize_best()"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "32624cbe",
   "metadata": {},
   "source": [
    "## What's next?\n",
    "\n",
    "Starting from the above version , you should code [the TD3\n",
    "algorithm](http://proceedings.mlr.press/v80/fujimoto18a/fujimoto18a.pdf).\n",
    "\n",
    "For that, you need to use two critics (and two target critics) and always take\n",
    "the minimum output between the two when you ask for the Q-value of a (state,\n",
    "action) pair.\n",
    "\n",
    "In more detail, you have to do the following:\n",
    "- replace the single critic and corresponding target critic with two critics\n",
    "  and target critics (name them ```critic_1, critic_2, target_critic_1,\n",
    "  target_critic_2```)\n",
    "- get the q-values and target q-values corresponding to all these critics.\n",
    "- then the target q-values you should consider to update the critic should be\n",
    "  the min over the target q-values at each step (use ```torch.min(...)``` to\n",
    "  get this min over a sequence of data).\n",
    "- to update the actor, do it with the q-values of an arbitrarily chosen\n",
    "  critic, e.g. critic_1."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "08834af0",
   "metadata": {
    "lines_to_next_cell": 1
   },
   "outputs": [],
   "source": [
    "class TD3(EpochBasedAlgo):\n",
    "    def __init__(self, cfg):\n",
    "        super().__init__(cfg)\n",
    "\n",
    "        # Define the agents and optimizers for TD3\n",
    "\n",
    "        assert False, 'Not implemented yet'\n",
    "\n",
    "\n",
    "\n",
    "\n",
    "def run_td3(td3: TD3):\n",
    "    for rb in iter_replay_buffers(td3):\n",
    "        rb_workspace = rb.get_shuffled(td3.cfg.algorithm.batch_size)\n",
    "\n",
    "        # Implement the learning loop\n",
    "\n",
    "        assert False, 'Not implemented yet'\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "b0a11029",
   "metadata": {
    "lines_to_next_cell": 2
   },
   "outputs": [],
   "source": [
    "params={\n",
    "  \"save_best\": False,\n",
    "  \"base_dir\": \"${gym_env.env_name}/td3-S${algorithm.seed}_${current_time:}\",\n",
    "  \"collect_stats\": True,\n",
    "  # Set to true to have an insight on the learned policy\n",
    "  # (but slows down the evaluation a lot!)\n",
    "  \"plot_agents\": True,\n",
    "  \"logger\":{\n",
    "    \"classname\": \"bbrl.utils.logger.TFLogger\",\n",
    "    \"cache_size\": 10000,\n",
    "    \"every_n_seconds\": 10,\n",
    "    \"verbose\": False,    \n",
    "    },\n",
    "\n",
    "  \"algorithm\":{\n",
    "    \"seed\": 1,\n",
    "    \"max_grad_norm\": 0.5,\n",
    "    \"epsilon\": 0.02,\n",
    "    \"n_envs\": 1,\n",
    "    \"n_steps\": 100,\n",
    "    \"nb_evals\": 10,\n",
    "    \"discount_factor\": 0.98,\n",
    "    \"buffer_size\": 2e5,\n",
    "    \"batch_size\": 64,\n",
    "    \"tau_target\": 0.05,\n",
    "\n",
    "    \"eval_interval\": 2_000,\n",
    "    \"max_epochs\": 11_000,\n",
    "    # Minimum number of transitions before learning starts\n",
    "    \"learning_starts\": 10000,\n",
    "    \"action_noise\": 0.1,\n",
    "    \"architecture\":{\n",
    "        \"actor_hidden_size\": [400, 300],\n",
    "        \"critic_hidden_size\": [400, 300],\n",
    "        },\n",
    "  },\n",
    "  \"gym_env\":{\n",
    "    \"env_name\": \"Pendulum-v1\",\n",
    "  },\n",
    "  \"actor_optimizer\":{\n",
    "    \"classname\": \"torch.optim.Adam\",\n",
    "    \"lr\": 1e-3,\n",
    "  },\n",
    "  \"critic_optimizer\":{\n",
    "    \"classname\": \"torch.optim.Adam\",\n",
    "    \"lr\": 1e-3,\n",
    "  }\n",
    "}\n",
    "\n",
    "td3 = TD3(OmegaConf.create(params))\n",
    "run_td3(td3)\n",
    "td3.visualize_best()"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "5e22db8c",
   "metadata": {},
   "source": [
    "## Experimental comparison\n",
    "\n",
    "Take an environment where the over-estimation bias may matter, and compare the\n",
    "performance of DDPG and TD3. Visualize the Q-value long before convergence to\n",
    "see whether indeed DDPG overestimates the Q-values with respect to TD3."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "760cc858",
   "metadata": {},
   "outputs": [],
   "source": [
    "from bbrl.stats import WelchTTest\n",
    "\n",
    "WelchTTest().plot(torch.stack(ddpg.eval_rewards), torch.stack(td3.eval_rewards), legends=\"ddpg/td3\", save=False)"
   ]
  }
 ],
 "metadata": {
  "jupytext": {
   "cell_markers": "\"\"\""
  },
  "kernelspec": {
   "display_name": "Python 3 (ipykernel)",
   "language": "python",
   "name": "python3"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}
